# frozen_string_literal: true

require 'discordrb/webhooks/version'
require 'discordrb/webhooks/embeds'
require 'discordrb/webhooks/client'
require 'discordrb/webhooks/builder'
require 'discordrb/webhooks/view'
require 'discordrb/webhooks/modal'

module Discordrb
  # Webhook client
  module Webhooks
  end
end
